#include "../include/ServerUtils.h"
#include "../include/WordUtils.h"
#include "../include/DataType.h"
#include "../include/FileUtils.h"

#include <string>
#include <utility>
#include <vector>
#include <iostream>
#include <sstream>
#include <unistd.h>
#include <netinet/in.h>
#include <thread>
#include <chrono>
#include <cstring>
#include <random>
#include <arpa/inet.h>
#include <sys/socket.h>


vector<string> css;
vector<string> images;
vector<unordered_map<string, unordered_map<string, int>>> all_lists;
shared_ptr<ConcurrentQueue<Track>> tracks;

void [[noreturn]] ServerUtils::serve(shared_ptr<ConcurrentQueue<Track>> cq_track) {
    tracks = std::move(cq_track);
    css = WordUtils::load_css("/home/skingging/Documents/Projects/CPP/AI-Tarpit-Reimagined/content/style.css");
    images = FileUtils::get_image_list("/home/skingging/Documents/Projects/CPP/AI-Tarpit-Reimagined/content/");
    const vector<string> words = FileUtils::get_wordlists("/home/skingging/Documents/Projects/CPP/AI-Tarpit-Reimagined/wordlist/");

    for (const auto& word : words) {
        all_lists.push_back(WordUtils::load_data(word));
    }

    // server_fd is a file descriptor.
    const int server_fd = socket(AF_INET, SOCK_STREAM, 0);
    
    sockaddr_in addr{};
    addr.sin_family = AF_INET;
    addr.sin_port = htons(8888);
    addr.sin_addr.s_addr = INADDR_ANY;

    if (bind(server_fd, reinterpret_cast<sockaddr *>(&addr), sizeof(addr)) != 0) {
        cout << "Please wait for the pipe to close.";
        return;
    }

    listen(server_fd, 50);
    
    cout << "Server is running on http://localhost:8888 \n";
    
    while (true) {
        int client_fd = accept(server_fd, nullptr, nullptr);

        if (client_fd == -1) continue;
        //cerr << client_fd << endl;
        thread(process_request, client_fd).detach();
    }
    
    close(server_fd);
}

void ServerUtils::process_request(const int client_fd) {
    string ip = get_ip(client_fd);

    char buffer[1024];
    const unsigned long bytes_received = recv(client_fd, buffer, sizeof(buffer) - 1, 0);

    string url;
    string user_agent;

    if (bytes_received > 0) {
        buffer[bytes_received] = '\0';

        url = WordUtils::extract_url(string(buffer));
        user_agent = WordUtils::extract_user_agent(string(buffer));

        if (url.empty()) {
            close(client_fd);
            cerr << "AAAA \n";
        }
    } else {
        close(client_fd);
        cerr << "AAAA \n";
    }

    Track track;
    track.Ip = ip;
    track.UserAgent = user_agent;

    tracks->push(track);

    if (url == "/style.css") {
        // This sends the header, that instructs how the browser should interpret the data.
        send_header(client_fd, data_type::CSS);
        send_chunked_css(client_fd);
    }

    else if (url == "/favicon.png") {
        send_image(client_fd, "/home/skingging/Documents/Projects/CPP/AI-Tarpit-Reimagined/a.png", image_type::PNG);
    }

    else if (WordUtils::contains_image(url)) {
        string p = "/home/skingging/Documents/Projects/CPP/AI-Tarpit-Reimagined/content/images/";
        p += WordUtils::extract_image_name(url);
        send_image(client_fd, p, image_type::AVIF);
    }

    else {
        const unsigned long hash3 = WordUtils::fnv1aHash(url);

        send_header(client_fd, HTML);
        send_chunked_html(client_fd, hash3);
    }

    // Send final zero-length chunk to end the response
    send(client_fd, "0\r\n\r\n", 5, 0);
    
    close(client_fd);
}

void ServerUtils::send_header(const int client_fd, const data_type type) {
    if (type == HTML) {
        send_data(client_fd, HTML_RESPONSE_HEADER);
    }
    
    else if (type == CSS) {
        send_data(client_fd, CSS_RESPONSE_HEADER);
    }
}

void ServerUtils::send_chunked_html(const int client_fd, const size_t hash) {
    const string hashes = to_string(hash);

    unsigned short itr = 0;
    const unsigned short end = hashes.size();

    send_data(client_fd, HTML_BEGINNING);
    send_data(client_fd, HTML_NAV);
    send_data(client_fd, HTML_MAIN_1);

    minstd_rand generator(hash);
    uniform_int_distribution<unsigned short> distribution_1(0, end - 1);
    uniform_int_distribution<unsigned short> distribution_2(0, images.size() - 1);
    uniform_int_distribution<unsigned short> distribution_3(0, 8); // There is only 9 wordlists.
    uniform_int_distribution<unsigned short> distribution_4(0, end - 2);

    const int link = distribution_1(generator);
    const int image = distribution_2(generator);
    const int l = distribution_3(generator);
    const int img = distribution_4(generator);

    while (itr < end) {
        send_data(client_fd, WordUtils::create_tag(all_lists[l], hashes[itr]));

        if (itr == link) {
            send_data(client_fd, WordUtils::create_link(all_lists[l], hash));
        }

        if (itr == img) {
            send_data(client_fd, WordUtils::create_image(images[image]));
        }

        //this_thread::sleep_for(chrono::milliseconds(75));

        itr++;
    }

    send_data(client_fd, HTML_MAIN_2);
    send_data(client_fd, HTML_FOOTER);
    send_data(client_fd, HTML_END);
}

void ServerUtils::send_chunked_css(const int client_fd) {
    for (const auto & cs : css) {
        send_data(client_fd, cs);

        this_thread::sleep_for(chrono::milliseconds(25));
    }
}

void ServerUtils::send_image(const int client_fd, const string& path, const image_type type) {
    const vector<unsigned char> image = FileUtils::open_image(path);

    std::ostringstream oss;
    oss << "HTTP/1.1 200 OK\r\n";
    if (type == PNG) {
        oss << "Content-Type: image/png\r\n";
    }

    if (type == AVIF) {
        oss << "Content-Type: image/avif\r\n";
    }

    oss << "Content-Length: " << image.size() << "\r\n";
    oss << "\r\n";

    send(client_fd, oss.str().c_str(), oss.str().size(), 0);

    send(client_fd, reinterpret_cast<const char*>(image.data()), image.size(), 0);
    //send_all(client_fd,reinterpret_cast<const char*>(image.data()), image.size());
}

size_t ServerUtils::send_all(const int client_fd, const char *data, const size_t length) {
    size_t total_sent = 0;
    while (total_sent < length) {
        const size_t sent = send(client_fd, data + total_sent, length - total_sent, 0);
        if (sent <= 0) {
            return sent;  // Error or connection closed
        }

        total_sent += sent;
    }

    return total_sent;
}

void ServerUtils::send_data(const int client_fd, const string& data) {
    ostringstream oss;

    oss << hex << data.size() << "\r\n" << data << "\r\n";

    send(client_fd, oss.str().c_str(), oss.str().size(), 0);
}

string ServerUtils::get_ip(const int client_fd) {
    sockaddr_storage addr{};
    socklen_t len = sizeof(addr);
    char buf[INET6_ADDRSTRLEN];

    if (getpeername(client_fd, reinterpret_cast<sockaddr *>(&addr), &len) == -1) {
        perror("getpeername");
        return "";
    }

    if (addr.ss_family == AF_INET) {
        // IPv4
        const sockaddr_in* s = reinterpret_cast<sockaddr_in *>(&addr);
        inet_ntop(AF_INET, &s->sin_addr, buf, sizeof(buf));
    } else {
        return "Unknown address family";
    }

    return buf;
}